/*******************************************************************************

NAME            pm_map.c
VERSION         %version: 3 %
UPDATE DATE     %date_modified: Tue Feb 24 11:51:06 2015 %
PROGRAMMER      %created_by:    kmoyer %

    Copyright 2014-2017 NetApp, Inc.

DESCRIPTION: Grub2 module to detect how much memory a system has and read a
config file to determine how the memory should be distributed on a LEOS system
between the kernel and EOS.

NOTES:
Config file is defined as follows:
  Lines beginning with '#' are comments.
  All other lines must contain 3 values, separated by "|".
    The first value is the memory size in GB.
    The second value is a pair indicating CBA memory size range.
    The third is the memory parameter string used to boot linux.
  If there are fewer than 3 values on a line,
    the line is ignored.
  If there are more than 3 values on a line,
    the extra values will be ignored.
  If: ACTUAL_MEM_SIZE == MEMSIZE  && CBA request is withing range
  Then: Use the corresponding memory parameter boot string.

LICENSE:
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************/


/***  INCLUDES  ***/
#include <config.h>
#include <config-util.h>

#include <grub/dl.h>
#include <grub/env.h>
#include <grub/file.h>
#include <grub/misc.h>
#include <grub/mm.h>
#include <grub/command.h>
#include <grub/i18n.h>
#include <grub/memory.h>

/***  CONSTANT DEFINITIONS  ***/
GRUB_MOD_LICENSE ("GPLv3+");
GRUB_MOD_DUAL_LICENSE ("All Rights Reserved");


/***  MACRO DEFINITIONS  ***/
#define DEFAULT_DIRECTORY  "/" GRUB_BOOT_DIR_NAME
#define CONFIG_FILE_NAME   "pm.map"
#define CONFIG_FILE        DEFAULT_DIRECTORY "/" CONFIG_FILE_NAME

#define DBG_PRT(arg...)  do{ if(enableDbgPrints) grub_printf(arg); } while (0)

/* Define constant for 1 GB */
#define GB (grub_uint64_t)(1024*1024*1024)

/* Round input value to nearest 1GB and then return as number of GB */
#define CONVERT_TO_GB(x)  (unsigned long)((x + (GB - 1)) / GB)

/***  TYPE DEFINITIONS  ***/
typedef struct {
    char*       start;
    char*       end;
    char*       cur;
    grub_off_t  len;
} ConfigFile;


/***  LOCALS  ***/
static int enableDbgPrints=0;

/***  PROCEDURES  ***/

/**
 * Takes a pointer and returns a pointer to the next new line (or NULL if there
 * are no new lines remaining).
 */
static char*
getNextString(ConfigFile* configFile)
{
    char* cur = configFile->cur + grub_strlen(configFile->cur);

    while (cur < configFile->end && *cur == '\0')
        cur++;

    configFile->cur = cur;

    if (cur < configFile->end)
        return cur;

    return 0;
}

/**
 * Free up memory allocated when file was open and read
 */
static void
configFileFree(ConfigFile* configFile)
{
    grub_free(configFile->start);
    configFile->start = 0;
}


/**
 * Read in the entire configuration file and store in structure.
 * Also convert all newlines to string terminators so that it is easier to parse.
 */
static int
readConfigFile(ConfigFile *configFile)
{
    grub_file_t  file;
    char        *start;

    file = grub_file_open(CONFIG_FILE);
    if (!file)
    {
        grub_printf("Could not open file: %s\n", CONFIG_FILE);
        return 0;
    }

    configFile->start = (char *)grub_zalloc(file->size);
    if (!configFile->start)
    {
        grub_file_close(file);
        return 0;
    }

    configFile->len = grub_file_read(file, configFile->start, file->size);
    if (configFile->len != file->size)
    {
        configFileFree(configFile);
        grub_file_close(file);
        return 0;
    }
    grub_file_close(file);

    configFile->cur = configFile->start;
    configFile->end = configFile->start + configFile->len;
    start = configFile->start;

    /* Scrub the file to get rid of newlines, replacing them with null terminator so that */
    /* parsing is easier */
    while (start < configFile->end)
    {
        if (*start == '\r' || *start == '\n')
            *start = '\0';
        start++;
    }

    return 1;
}


/**
 * Takes the actual/physical mem size and looks for a corresponding memory
 * mapping size in the config file. Returns 0 if no valid config line found.
 * If it returns mapping, that memory will be allocated and needs to be freed
 * by the caller.
 */
static char*
getMemConfiguration(unsigned long actualMemSize, unsigned long cbaMemReq)
{
    ConfigFile configFile;
    char *cur;
    char *retbuf = 0;
    unsigned long memsize;
    unsigned long cba_max;
    unsigned long cba_min;

    DBG_PRT("actualMemSize: %lu\n", actualMemSize);

    if (!readConfigFile(&configFile))
    {
        return 0;
    }

    cur = getNextString(&configFile);

    while (cur != NULL && *cur != '\0')
    {
        /* Advance to the next non-white space character */
        while((*cur != '\0') && grub_isspace(*cur))
        {
            cur++;
        }

        /* If comment line, then skip to next line in file */
        if ( *cur == '#')
        {
            cur = getNextString(&configFile);
            continue;
        }

        /* Read in UL hex value, advance 'cur' */
        memsize = grub_strtoul(cur, &cur, 10);
        DBG_PRT("memsize: %lu \n", memsize);
        if (memsize == 0 || memsize != actualMemSize || cur == NULL || *cur == '\0')
        {
            cur = getNextString(&configFile);
            continue;
        }
        if (*cur == '|')
             cur++;

        /* Get CBA range indicated for this line (values are inclusive) */
        cba_min = grub_strtoul(cur, &cur, 10);
        cba_max = grub_strtoul(cur, &cur, 10);

        /* Use the 0 0 configuration by default, then use the CBA configuration if it is
           present. */
        if ((cba_min == 0 && cba_max == 0)
            || (cbaMemReq >= cba_min && cbaMemReq <= cba_max))
        {
            if (*cur == '|')
                 cur++;

            if (cur != NULL && *cur != '\0')
            {
               grub_free(retbuf);
               retbuf = grub_strdup(cur);
            }
        }

        cur = getNextString(&configFile);
    }

    configFileFree(&configFile);
    return retbuf;
}

/**
 * Called by grub_machine_mmap_iterate() for each region
 * Simply add up the available region sizes
 */
static int
avail_mem_hook(grub_uint64_t addr __attribute__ ((unused)),
               grub_uint64_t size,
               grub_memory_type_t type,
               void *data)
{
   grub_uint64_t *total_mem = data;

   if (type == GRUB_MEMORY_AVAILABLE)
   {
       (*total_mem) += size;
   }

   return GRUB_ERR_NONE;
}

/**
 * The guts of the 'pm_map' command.
 * Determine the amount of memory available on the system and then check the
 * config file to see if it's a supported range. If so, set the desired env
 * variable to the appropriate string.
 */
static grub_err_t
grub_cmd_pm_map (grub_command_t cmd __attribute__ ((unused)),
                int argc,
                char *argv[])
{
    char *buf;
    grub_uint64_t fullMemCap = 0;
    unsigned long cbaMemReqGb = 0;
    char *varName;
    char *cbaReqString;

    if (argc==0 || argc > 2)
    {
        return grub_error(GRUB_ERR_BAD_ARGUMENT, N_("one or two arguments expected"));
    }

    /* If "DEBUG" env var exists, enable prints */
    enableDbgPrints = grub_env_get("DEBUG") ? 1 : 0;

    varName = argv[0];
    DBG_PRT("variable: %s\n", varName);

    if (argc == 2)
    {
        cbaReqString = argv[1];
        DBG_PRT("CBA req: %s\n", cbaReqString);
    }
    else
    {
        cbaReqString = (char*) "0";
        DBG_PRT("CBA def: %s \n", cbaReqString);
    }
    cbaMemReqGb = grub_strtoul(cbaReqString, NULL, 10);

#ifndef GRUB_MACHINE_EMU
    grub_machine_mmap_iterate(avail_mem_hook,&fullMemCap);
#endif

    buf = getMemConfiguration(CONVERT_TO_GB(fullMemCap), cbaMemReqGb);
    if (buf)
    {
        DBG_PRT("Set %s=%s\n", varName, buf);
        grub_env_set(varName, buf);
    }
    else
    {
        grub_printf("\nCould not determine system resource allocation: %s=(null)\n", varName);
        grub_env_set(varName, "");
    }

    grub_free(buf);
    DBG_PRT("pmm end\\n");

    return GRUB_ERR_NONE;
}


/* Register/unregister our command with module init/exit */
static grub_command_t cmd;

GRUB_MOD_INIT(pm_map)
{
    cmd = grub_register_command("pm_map", grub_cmd_pm_map, N_("ENV_VAR CBA_REQ_IN_GB"),
                                N_("Remaps MP table to allocate reserved block for protected memory."));
}

GRUB_MOD_FINI(pm_map)
{
    if (cmd)
    {
        grub_unregister_command(cmd);
        cmd = 0;
    }
}

