/*******************************************************************************

NAME            var_read.c
SUMMARY         %description%
VERSION         %version:   3 %
UPDATE DATE     %date_modified: Thu Apr 14 16:00:21 2016 %
PROGRAMMER      %created_by: bradfort %

    Copyright 2014-2016 NetApp, Inc.

DESCRIPTION: This grub module reads the given file content from a persistent
             location and sets the grub environment variable with the value.

LICENSE:
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include <config-util.h>
#include <grub/env.h>
#include <grub/file.h>
#include <grub/dl.h>
#include <grub/command.h>

GRUB_MOD_LICENSE ("GPLv3+");
GRUB_MOD_DUAL_LICENSE ("All Rights Reserved");

#define CONFIG_DIR  "(lvm/system-data)/earlypersist"

#define DBG_PRT(arg...)    \
{                          \
    if (enableDbgPrints)   \
        grub_printf(arg);  \
}

static int enableDbgPrints = 0;

/**
 * This function opens a file with the given variable name,
 * reads its contents and returns it.
 */
static grub_err_t
grub_read_var(char *varName, char *buf, int len)
{
    char config_file[128];
    grub_file_t file;
    grub_size_t nchar;

    grub_snprintf(config_file, sizeof(config_file), "%s/%s", CONFIG_DIR, varName);
    DBG_PRT("config file: %s\n", config_file);

    file = grub_file_open(config_file);
    if (!file)
    {
        grub_printf("Could not open file: %s\n", config_file);
        return grub_errno;
    }

    nchar = grub_file_read(file, buf, len-1);
    if (nchar != file->size)
    {
        grub_printf("grub file read failed.\n");
        grub_file_close(file);
        return grub_errno;
    }

    grub_file_close(file);

    DBG_PRT("Length buffer read is %d\n", len);

    /* Null terminate the string */
    buf[nchar] = 0;

    return GRUB_ERR_NONE;
}

/**
 * This function implements the veos_read grub command.
 * It reads the variable value from the config file and creates/sets
 * the grub environment variable with the value. If it can't read
 * the config file for some reason, it will use the default
 * value passed to it.
 */
static grub_err_t
grub_cmd_veos_read(grub_command_t cmd __attribute__ ((unused)),
                int argc,
                char *argv[])
{
    char value[128];
    char *varName;
    char *defaultValue;

    if (argc != 2)
    {
        return grub_error(GRUB_ERR_BAD_ARGUMENT, N_("two arguments are expected"));
    }

    /* If "DEBUG" env var exists, enable prints */
    enableDbgPrints = grub_env_get("DEBUG") ? 1 : 0;

    varName = argv[0];
    defaultValue = argv[1];

    DBG_PRT("variable: %s\n", varName);
    DBG_PRT("default value: %s\n", defaultValue);

    if (grub_read_var(varName, value, sizeof(value)))
    {
        /* If it failed to read the value from config file, then assign the default value */
        grub_strncpy(value, defaultValue, sizeof(value));

        /* Ignore the error to avoid waiting for user response during boot */
        grub_errno = 0;
    }

    DBG_PRT("Set %s=%s", varName, value);
    grub_env_set(varName, value);

    return GRUB_ERR_NONE;
}

static grub_command_t cmd;

/* Register/unregister our command with module init/exit */
GRUB_MOD_INIT(veos_read)
{
    cmd = grub_register_command("veos_read", grub_cmd_veos_read, N_("ENV_VAR"),
                                N_("Reads kernel variable values from vEOS persist file."));
}

GRUB_MOD_FINI(veos_read)
{
    if (cmd)
    {
        grub_unregister_command(cmd);
        cmd = 0;
    }
}
